/* * C o p y r i g h t *(also read COPYING)* * * * * * * * * * * * * * * * * * *
 *                                                                             *
 *  Copyright (C) 2002  <Philip Van Hoof>                                      *
 *  Copyright (C) 2003  <Mark Finlay>                                	       *
 *                                                                             *
 *  This program is free software; you can redistribute it and/or modify       *
 *  it under the terms of the GNU General Public License as published by       *
 *  the Free Software Foundation; either version 2 of the License, or          *
 *  (at your option) any later version.                                        *
 *                                                                             *
 *  This program is distributed in the hope that it will be useful,            *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 *  GNU General Public License for more details.                               *
 *                                                                             *
 *  You should have received a copy of the GNU General Public License          *
 *  along with this program; if not, write to the Free Software                *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 *                                                                             *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */ 

#include "xsu.h"

#ifdef CHECK_SU_OUTPUT
	static gint
	exec_su_failed (gpointer user_data)
	{
		#ifdef DEBUG
    			printf("static gint exec_su_failed (gpointer user_data)\n");
		#endif
		gnome_dialog_run_and_close (GNOME_DIALOG (gnome_error_dialog (
		_("Timeout when trying to execute 'su'."))));
		gtk_exit (EXIT_ERROR);
		return FALSE;
	}
#endif


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
*	
*			Su Error Message
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */	
	
static void
su_has_died (GtkWidget *term, gpointer user_data)
{
	gchar *buf;
	gint len;

	#ifdef DEBUG
		printf("static void su_has_died (GtkWidget *widget, gpointer user_data)\n");
	#endif
	buf = zvt_term_get_buffer (ZVT_TERM (term), &len, VT_SELTYPE_CHAR, 3, 1, 200, 1);

	if (strlen(buf) > 1) 
	{	
		gchar *display;

		display = g_strdup_printf("Your command outputed the following:: %s ", buf);
		// gnome_dialog_run_and_close (GNOME_DIALOG (gnome_error_dialog (_(display))));
        printf(display);
        printf("\n");
		gtk_exit(EXIT_ERROR);
	} else { 
		gtk_exit(EXIT_CORRECT);
	}
		while (gtk_events_pending ()) gtk_main_iteration ();
		gtk_main_quit ();
	}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
*	
*			Perform Su Opperation
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
	
void xsu_perform()
{ 
/* 
	Some code from gnomesu, thanks to Hongli Lai <hongli@telekabel.nl> 
*/
	GtkWidget *term;
	guint su_d13;
	int i; /* No gint. We will use this in a non-gnome thread */
	gchar* password_return;
	const gchar* username;
	const gchar* password;
	const gchar* command;
	gchar *buffer;

#ifdef CHECK_SU_OUTPUT
	guint timeout;
#endif

#ifdef DEBUG
		printf("void xsu_perform()\n");
#endif

	gtk_xsu_window = glade_xml_get_widget (xml, "gtk_xsu_window");
	gtk_widget_hide (gtk_xsu_window);

	/** G_CONST_RETURN see http://developer.gnome.org/doc/API/2.0/gtk/GtkEntry.html#gtk-entry-get-text */
	username = gtk_entry_get_text (GTK_ENTRY (gtk_user_textbox));
	password = gtk_entry_get_text (GTK_ENTRY (gtk_password_textbox));
	command = gtk_entry_get_text (GTK_ENTRY (gtk_command_textbox));

	if (password == NULL)
		return;

	term = zvt_term_new ();
	i = zvt_term_forkpty (ZVT_TERM (term), 
		ZVT_TERM_DO_UTMP_LOG | ZVT_TERM_DO_WTMP_LOG);

	su_d13 = gtk_signal_connect (GTK_OBJECT (term), "child_died", 
		GTK_SIGNAL_FUNC (su_has_died), NULL);

	if (i == 0)
	{

	/* 	execlp() has been replaced with execl()
		This also fixes the problem that the zvt
		widget could not get the child_died signal. */
	if (set_display_env) 
	{
		buffer = g_strdup_printf("%s%s;%s", SET_DISPL_ENV, displ_host, command);
		g_free(displ_host);
	} 
	else 
	{
		buffer = g_strdup_printf("%s", command);
	}
	
	/* xsu 0.2.2 *
		Security issue noticed by Havoc:
		Security fix, execlp exports the PATH which is a security problem. SU_PATH contains
		the full path of the su binary found by the configure script.
	
		buffer with " and/or ' gives some problems. 
		buffer = g_strdup_printf("%s - %s -c %s",SU_PATH, username, buffer); */

		/* running 'su - username -c command' */
	execl(SU_PATH, "-", username, "-c", buffer, NULL);
		/* We are not a gnome-application anymore but under control of su */
		
		/*username=g_strdup_printf("%s@localhost");
		  execl("/usr/bin/ssh", username, "-s", buffer, NULL);*/
		/* I tried using the SSH client for this. But it then gives an error
			that it has been put in the background after auth. */
		
	perror (NULL);
	exit (EXIT_ERROR);

	} else if (i == -1) {
		perror (NULL);
		gtk_exit (EXIT_ERROR);
	}
	
	/* xsu 0.2.4: Temporary fix */
#ifdef CHECK_SU_OUTPUT
		timeout = gtk_timeout_add (SU_DELAY, exec_su_failed, NULL);
		while (1)
		{
			char *buf;
			int len;

			buf = zvt_term_get_buffer (ZVT_TERM (term), &len,
				VT_SELTYPE_CHAR, 0, 0, strlen (SU_PWD_OUT), 0);
			if (strcmp (buf, SU_PWD_OUT) == 0)
			{
				gtk_timeout_remove (timeout);
				g_free (buf);
				break;
			}

			g_free (buf);
			while (gtk_events_pending ()) gtk_main_iteration ();
			usleep (5);
		}
#endif

	usleep(200);
	password_return = g_strdup_printf ("%s\n", password);
		
	/* Clear the password to the best of my ability. If you can think of a better way to
	do this then please e-mail me. */
	gtk_entry_set_text(GTK_ENTRY(gtk_password_textbox),"");
	memset ((gchar*)password, 0, strlen (password));
	zvt_term_writechild (ZVT_TERM (term), password_return, strlen (password_return));
	memset (password_return, 0, strlen (password_return));
	g_free (password_return);

	gtk_signal_disconnect (GTK_OBJECT (term), su_d13);
	
	gtk_signal_connect (GTK_OBJECT (term), "child_died",
		GTK_SIGNAL_FUNC (su_has_died), NULL);


	gtk_main();
	return;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
*	
*				Signal Functions
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void on_gtk_user_textbox_activate (GtkButton *button, gpointer user_data)
{
#ifdef DEBUG
    printf("on_gtk_user_textbox_activate (GtkButton *button, gpointer user_data)\n");
#endif
    gtk_widget_grab_focus (gtk_password_textbox);
}

void on_gtk_command_textbox_activate (GtkButton *button, gpointer user_data)
{
#ifdef DEBUG
    printf("on_gtk_command_textbox_activate (GtkButton *button, gpointer user_data)\n");
#endif
	if (GTK_WIDGET_VISIBLE( glade_xml_get_widget (xml, "gtk_user_textbox") ))
		gtk_widget_grab_focus ( glade_xml_get_widget (xml, "gtk_user_textbox") );
	else
		gtk_widget_grab_focus (gtk_password_textbox);
}

void on_gtk_cancel_button_clicked (GtkButton *button, gpointer user_data)
{
#ifdef DEBUG
    printf("on_gtk_cancel_button_clicked (GtkButton *button, gpointer user_data)\n");
#endif
    exit(EXIT_ERROR);
}

void quit_program(){
    gtk_main_quit();
    exit(EXIT_CORRECT);
}

void on_gtk_ignore_button_clicked (GtkButton *button, gpointer user_data)
{
	gchar* command;

#ifdef DEBUG
    printf("on_gtk_ignore_button_clicked (GtkButton *button, gpointer user_data)\n");
#endif
	
	gtk_command_textbox = glade_xml_get_widget (xml, "gtk_command_textbox");
	command = (gchar*)gtk_entry_get_text (GTK_ENTRY (gtk_command_textbox));
	execlp(command, command, NULL);

}

void on_gtk_ok_button_clicked (GtkButton *button, gpointer user_data)
{
	gchar* password;

#ifdef DEBUG
    printf("on_gtk_ok_button_clicked (GtkButton *button, gpointer user_data)\n");
#endif
	// If there is no password then run the program as normal user.
	gtk_password_textbox = glade_xml_get_widget (xml, "gtk_password_textbox");
	password = (gchar*)gtk_entry_get_text (GTK_ENTRY (gtk_password_textbox));
	
	if (!strcmp ("", password))
	{
		on_gtk_ignore_button_clicked (NULL, NULL);
	}
	else
	{
		xsu_perform();
	}
	memset ((gchar*)password, 0, strlen (password));
}

void on_gtk_password_textbox_activate (GtkButton *button, gpointer user_data)
{
#ifdef DEBUG
    printf("on_gtk_user_textbox_activate (GtkButton *button, gpointer user_data)\n");
#endif
    on_gtk_ok_button_clicked(NULL,NULL);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
*	
*							Main Function
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int main (int argc, char *argv[])
{
	GtkWidget*	gtk_xsu_window;
	gchar*		arg_command;
	gchar*		arg_username=NULL;
	gint 	  x;
	gboolean command_in=FALSE;
	gboolean username_in=FALSE;
	gboolean endofargs=FALSE;
	gboolean first=TRUE;
	gchar* glade_common_path;

    /*	Nope..no popt, maybe next time. Oh and getopt stuff is not
		the same in *BSD as with GNU getopt, so I don't immedialty 
		feel like wanting to change this option parsing :-\. 

		Some problems about parsing the arguments have been filed by
		Jeff Licquia <licquia@debian.org>. I added the "--" option
		which stops parsing. (Only " and \ have to be escaped of course,
		this depends on your shell)
	*/

	/* Initialize the i18n stuff */
	// gcc says that these statments have no effect
	 //bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
	 //textdomain (PACKAGE);
  
  
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
*	
*			Command line Arguments
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
	for (x=0; x<argc; x++)
	{
		/* Argument "--" stops argument parsing and collects all arguments
		after the "--" argument. It will put all these arguments in one
		string and use this string in arg_command. */
		if (!strcmp (argv[x], "--")) endofargs=TRUE;
			
		if (endofargs) {
			gchar *buffer;
			buffer = g_strdup(argv[x+1]);
			if (buffer) {
				if (first) {
					arg_command = g_strdup_printf("%s",buffer);
					first=FALSE;
				} else {
					arg_command = g_strdup_printf("%s %s", arg_command, buffer);
				}
				continue;
			} else {
				command_in=TRUE;
				break;
			}
		}
		
    
		if ((!strcmp (argv[x], "-u")) || (!strcmp (argv[x], "--username"))) // there was a -u statment
		{
			arg_username = "";
			if (argv[x+1] != NULL) // there is a username value specified with -u
			{	

				arg_username = g_strdup(argv[x+1]);
				username_in=TRUE;
				
				if (!message_in) // set a custom message if none was set by the user.
				{
					arg_message = g_strdup_printf("<b>This dialog will run a program as the user %s.</b>\n Please enter that user's password and press OK to continue.", arg_username);
					message_in=TRUE;
				}
			}
			else // there was a -u but there was no username specified. We want to show the username box to let the user decide
			{
				username_in=FALSE;
				arg_username="";
			}
		}

		if(g_strdup(argv[1]))	// there was a first argument
		{
			/* As long as the first item is not a switch set the first item as the command */
			if ((strcmp(argv[1], "-u")) && (strcmp(argv[1], "--username")) && (strcmp(argv[1], "-m")) && (strcmp (argv[1], "-c")) && (strcmp (argv[1], "--command")) && (strcmp(argv[1], "--message")) && (strcmp(argv[1], "-t")) && (strcmp(argv[1], "--title")))
			{
				arg_command = g_strdup(argv[1]);
				command_in=TRUE;
			}		
		}

		// -c switch for backwards compatibility with xsu	
		if ((!strcmp (argv[x], "-c")) || (!strcmp (argv[x], "--command")))
		{
			if (argv[x+1] != NULL)
			{
				arg_command = g_strdup(argv[x+1]);
				command_in=TRUE;
			}
		}			

		if ((!strcmp (argv[x], "-m")) || (!strcmp (argv[x], "--message")))
		{
			if (argv[x+1] != NULL)
			{	
				int i;
				arg_message = g_strdup(argv[x+1]);
				for (i=0;i<strlen(arg_message);++i)
					if (arg_message[i]=='^') arg_message[i]='\n';
				for (i=0;i<strlen(arg_message);++i)
					if (arg_message[i]=='~') arg_message[i]='\t';
				message_in=TRUE;
			}
		}
/* xsu 0.1.5 *
	Option to load another icon

		if ((!strcmp (argv[x], "-i")) || (!strcmp (argv[x], "--icon")))
		{
			if (argv[x+1] != NULL)
			{
				arg_iconfile = g_strdup(argv[x+1]);
				icon_in=TRUE;
			}
		}
		*/	
		
		
		if ((!strcmp (argv[x], "-t")) || (!strcmp (argv[x], "--title")))
		{
			if (argv[x+1] != NULL)
			{
				arg_title = g_strdup(argv[x+1]);
				title_in=TRUE;
			}
		}

		/* Option to set the DISPLAY environment
  		This seems to be broken. I have no way to test as I'm not exactly sure what it's ment to do. */
		if ((!strcmp (argv[x], "-a")) || (!strcmp (argv[x], "--set-display")))
		{
			set_display_env=TRUE;
	    		if ((argv[x+1] != NULL) && ((argv[x+1])[0] != '-'))
	    		{ /* if there is a 2e argument and it does not begin with a dash */
					displ_host = g_strdup(argv[x+1]);
	    		} else {
					displ_host = g_strdup(":0");
	    		}
		}
		
		if ((!strcmp (argv[x], "-v")) || (!strcmp (argv[x], "--version")))
		{
			printf(NAME " " VERSION " by " AUTHOR "\n");
			exit(EXIT_CORRECT);
		}
 
		if ((!strcmp (argv[x], "-h")) || (!strcmp (argv[x], "--help"))/* || (!g_strdup(argv[1]))*/)
		{
			printf(NAME " " VERSION " by " AUTHOR "\n\n");
			printf("Usage :\n");
			printf("\tgnomesu\t\t\t\t - You can specify the command\n\t\t\t\t\t   and password in the gui.\n");
			printf("\tgnomesu <command> <arguments>\t - You can specify the command and\n\t\t\t\t\t   arguments on the command line.\n");
			printf("\tgnomesu <arguments>\t\t - You can specify arguments on the\n\t\t\t\t\t   command line and choose the command\n\t\t\t\t\t   in the gui.\n");
			printf("\tgnomesu -u\t\t\t - You can specify command, user and\n\t\t\t\t\t   password in the gui.\n\n");
			
			printf("Optional Arguments :\n");
			printf("\t-u|--username \"USERNAME\"\n");
			printf("\t-m|--message \"MESSAGE\"\n");
			printf("\t-t|--title \"WINDOW TITLE\"\n");
			printf("\t-c|--command \"COMMAND\"\n");
			printf("\t-v|--version\n");
			printf("\t-h|--help\n\n");
//			printf("\t-a|--set-display [\"HOSTNAME\"]\n");
//			printf("\t-i|--icon \"FILENAME\"\n");

			printf("For MESSAGE :\n");
			printf("\t* You can use ^ if you want a newline in the message string.\n");
			printf("\t* You can use ~ if you want a tab in the message string.\n\n");  
			
			exit(EXIT_CORRECT);
		}

	}

	// Default to root if no username specified
	if (arg_username == NULL)
	{
		//Setting root as the default user as none is specified
		arg_username = "root";
		username_in=TRUE;
	}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
*	
*				Start Program
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
	
/* 	This one is to confuse gnome_init() actually. I don't want
	gnome to use the program arguments. And I am to lazy to learn
	how to use popt. Well.. if you are not..modify so that it uses
	popt and send me the patch :)*/

	argc=1;
	gnome_init (NAME, VERSION, argc, argv);
	gtk_init(&argc, &argv);
	
	// this is the old function used to draw the UI
	// TODO: Remove glade dep and use gtk again.
	//gtk_xsu_window = create_gtk_xsu_window ();
	
	// draw glade interface
	glade_common_path = g_strdup_printf ("%s/xsu.glade", PACKAGE_DATA_DIR);
	xml = glade_xml_new(glade_common_path, NULL, NULL);
	if(!xml) 
	{
         	g_warning("We could not load the glade interface file! Please check that you have xsu installed correctly");
		return 1;
	}
	g_free(glade_common_path);

    	glade_xml_signal_autoconnect(xml);
	gtk_xsu_window = glade_xml_get_widget (xml, "gtk_xsu_window");
	gtk_user_textbox = glade_xml_get_widget (xml, "gtk_user_textbox");
	gtk_password_textbox = glade_xml_get_widget (xml, "gtk_password_textbox");
	gtk_command_textbox = glade_xml_get_widget (xml, "gtk_command_textbox");
	gtk_txtuser_label = glade_xml_get_widget (xml, "gtk_txtuser_label");
	gtk_commandtxt_label = glade_xml_get_widget (xml, "gtk_commandtxt_label");
	gtk_text_label = glade_xml_get_widget (xml, "gtk_text_label");

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
*	Modify UI based on Command Line Agruments
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	if (username_in)
	{	
		gtk_entry_set_text(GTK_ENTRY(gtk_user_textbox),arg_username);
	}	
	else
	{
		gtk_widget_show(gtk_user_textbox);
		gtk_widget_show(gtk_txtuser_label);
		arg_message = "<b>This dialog will run a program as the specified user.</b>\n Please enter a username and that user's password and press OK to continue.";
		message_in=TRUE;
	}

	if (message_in)
	{
		gtk_label_set_markup(GTK_LABEL(gtk_text_label),arg_message);
	}
	
	if (command_in)
	{
		gtk_entry_set_text(GTK_ENTRY(gtk_command_textbox),arg_command);
		//gtk_entry_set_editable(GTK_ENTRY(gtk_command_textbox), FALSE);
		gtk_widget_grab_focus (gtk_password_textbox);
	} else {
		gtk_widget_grab_focus (gtk_command_textbox);
	}
	if (!username_in)
	{
		gtk_widget_grab_focus (gtk_user_textbox);
	}
	if (title_in)
 	{
		gtk_window_set_title (GTK_WINDOW (gtk_xsu_window), arg_title);
  	}
	
/* Option to load another icon
Broken ATM
	if (icon_in) {
		gtk_widget_destroy(gtk_keys_pixmap);
		//gtk_keys_pixmap = gnome_pixmap_new_from_file(arg_iconfile);
		gtk_keys_pixmap = gtk_image_new_from_file(arg_iconfile);
		
		gtk_widget_set_name (gtk_keys_pixmap, "gtk_keys_pixmap");
		gtk_widget_show (gtk_keys_pixmap);
		gtk_table_attach (GTK_TABLE (gtk_window_table), gtk_keys_pixmap, 0, 1, 0, 1,
					(GtkAttachOptions) (GTK_FILL),
					(GtkAttachOptions) (GTK_FILL), 0, 0);
	}
*/	
	


/*  Setting the window icon
FIXME: Find a better icon, check jimmac's collection
*/
	pixbuf = NULL;
	file = gnome_program_locate_file (NULL, GNOME_FILE_DOMAIN_PIXMAP,  "gnome-warning.png", TRUE, NULL);
	if (!file) {
		g_warning (G_STRLOC ": icon was not found cannot be found");
	} else {
		pixbuf = gdk_pixbuf_new_from_file (file, NULL);
		gtk_window_set_icon (GTK_WINDOW (gtk_xsu_window), pixbuf);
	}

	gtk_widget_show (gtk_xsu_window);
	if (SU_PATH[0]!='/') g_print("Warning! the path to your su binary does not start with a /.\nThis could mean that your system is compromised.\n");
	gtk_main ();    

	return 0;
}
